const std = @import("std");
const sdl = @import("sdl2");
const zlm = @import("zlm");
const gl = @import("zgl");
const m = zlm.SpecializeOn(f32);

const Self = @This();

window: sdl.Window,
context: sdl.gl.Context,
//color_loc: u32,
buffer: gl.Buffer,
color: [4]f32,

test_counter: usize = 0,

vbo: [max_objects]f32 = .{0.0} ** max_objects,
//colors: [max_objects][4]f32 = .{[4]f32{ 0.0, 0.0, 0.0, 1.0 }} ** max_objects,
vbo_index: usize = 0,

const max_objects: usize = 16384;

fn glGetProcAddress(p: []const u8, proc: [:0]const u8) ?*const anyopaque {
    _ = p;
    return sdl.c.SDL_GL_GetProcAddress(@ptrCast([*c]const u8, proc));
}

pub fn init() !Self {
    try sdl.init(.{ .video = true, .audio = true, .events = true });
    try sdl.ttf.init();

    try sdl.gl.setAttribute(.{ .context_major_version = 3 });
    try sdl.gl.setAttribute(.{ .context_minor_version = 3 });
    try sdl.gl.setAttribute(.{ .multisamplebuffers = true });
    try sdl.gl.setAttribute(.{ .multisamplesamples = 4 });

    const window = try sdl.createWindow(
        "USG",
        .{ .centered = {} },
        .{ .centered = {} },
        1280,
        720,
        .{ .context = .opengl, .vis = .shown },
    );

    const ctx = try sdl.gl.createContext(window);
    sdl.gl.setSwapInterval(.immediate) catch {
        std.debug.print("WARNING: Unable to configure the swap interval.\n", .{});
    };

    const proc: []const u8 = undefined;
    try gl.loadExtensions(proc, glGetProcAddress);

    // Shader stuff
    var mvp_loc: u32 = undefined;
    const program = gl.Program.create();
    {
        const vs = gl.Shader.create(.vertex);
        defer vs.delete();
        const fs = gl.Shader.create(.fragment);
        defer fs.delete();

        vs.source(1, &.{@embedFile("shaders/vector.vs")});
        fs.source(1, &.{@embedFile("shaders/fragment.fs")});

        vs.compile();
        fs.compile();

        program.attach(vs);
        defer program.detach(vs);
        program.attach(fs);
        defer program.detach(fs);

        program.link();

        mvp_loc = program.uniformLocation("mvp").?;
    }
    program.use();

    // Get the matrix that converts coordinates to 1:1 on the screen from the OpenGL default
    {
        const wsize = window.getSize();
        const xunit = @intToFloat(f32, wsize.width);
        const yunit = @intToFloat(f32, wsize.height);

        const ortho = m.Mat4.createOrthogonal(0.0, xunit, yunit, 0.0, 0.0, 100.0);

        gl.uniformMatrix4fv(mvp_loc, false, &.{ortho.fields});
    }

    var vertex_array = gl.VertexArray.gen();
    vertex_array.bind();

    const buf = gl.Buffer.gen();
    buf.bind(.array_buffer);

    gl.vertexAttribPointer(0, 2, .float, false, @sizeOf(f32) * 8, @sizeOf(f32) * 0);
    gl.enableVertexAttribArray(0);

    gl.vertexAttribPointer(1, 4, .float, false, @sizeOf(f32) * 8, @sizeOf(f32) * 2);
    gl.enableVertexAttribArray(1);

    gl.vertexAttribPointer(2, 2, .float, false, @sizeOf(f32) * 8, @sizeOf(f32) * 6);
    gl.enableVertexAttribArray(2);

    gl.clearColor(0.91, 0.85, 0.65, 1.00);

    // TODO: See if there's a more optimal solution to this, maybe activating only when necessary
    gl.enable(.blend);
    gl.blendFunc(.src_alpha, .one_minus_src_alpha);

    var text = Text.init("testing123");
    text.deinit();

    return Self{ .window = window, .context = ctx, .buffer = buf, .color = .{ 0, 0, 0, 0 } };
}

pub fn render(self: *Self) void {
    //std.debug.print("Number of rendered elements: {}\n", .{self.test_counter});
    self.test_counter = 0;

    self.buffer.data(f32, &self.vbo, .static_draw);

    gl.drawArrays(.triangles, 0, self.vbo_index);
    sdl.gl.swapWindow(self.window);
    gl.clear(.{ .color = true });

    self.vbo_index = 0;

    // Clear the vbo, this really shouldn't be necessary as the index is set to zero,
    // but otherwise it leads to bugs
    self.vbo = .{0.0} ** max_objects;
}

pub fn deinit(self: *Self) void {
    gl.disableVertexAttribArray(0);
    sdl.quit();
    sdl.ttf.quit();
    self.window.destroy();
    sdl.gl.deleteContext(self.context);
}

pub fn setColorF(self: *Self, r: f32, g: f32, b: f32, a: f32) void {
    self.*.color = .{ r, g, b, a };
}

pub fn setColor(self: *Self, r: i32, g: i32, b: i32, a: i32) void {
    self.*.color = .{
        @intToFloat(f32, r) / 255.0,
        @intToFloat(f32, g) / 255.0,
        @intToFloat(f32, b) / 255.0,
        @intToFloat(f32, a) / 255.0,
    };
}

pub fn drawRectangle(self: *Self, x: i32, y: i32, w: i32, h: i32) void {
    //const indices = [_]u8{ 0, 1, 1, 2, 2, 3, 3, 0 };
    //self.renderRectangle(x, y, w, h, .lines, &indices);
    _ = x;
    _ = y;
    _ = self;
    _ = w;
    _ = h;
}

pub fn fillRectangle(self: *Self, x: i32, y: i32, w: i32, h: i32) void {
    self.fillRectangleEx(x, y, w, h, 0);
}

pub fn fillRectangleEx(self: *Self, x: i32, y: i32, w: i32, h: i32, skew_x: i32) void {
    // TODO: Is this still necessary?
    self.test_counter += 1;

    var xf = @intToFloat(f32, x);
    var yf = @intToFloat(f32, y);
    var wf = @intToFloat(f32, w);
    var hf = @intToFloat(f32, h);

    const skew_x_offset = @intToFloat(f32, skew_x) * hf / 100;

    const i = self.vbo_index;
    const vertex_data = [_]f32{
        xf + skew_x_offset, yf, // up-left
        self.color[0], self.color[1], self.color[2], self.color[3],
        0.0, 0.0,
        xf + wf + skew_x_offset, yf, // up-right
        self.color[0], self.color[1], self.color[2], self.color[3],
        1.0, 0.0,
        xf - skew_x_offset, yf + hf, // down-left
        self.color[0], self.color[1], self.color[2], self.color[3],
        0.0, 1.0,
        xf + wf - skew_x_offset, yf + hf, // down-right
        self.color[0], self.color[1], self.color[2], self.color[3],
        1.0, 1.0,
        xf + wf + skew_x_offset, yf, // up-right
        self.color[0], self.color[1], self.color[2], self.color[3],
        1.0, 0.0,
        xf - skew_x_offset, yf + hf, // down-left
        self.color[0], self.color[1], self.color[2], self.color[3],
        0.0, 1.0,
    };
    std.mem.copy(f32, self.vbo[i .. i + vertex_data.len], &vertex_data);
    self.vbo_index += vertex_data.len;
}

fn renderRectangle(self: *Self, x: i32, y: i32, w: i32, h: i32) void {
    self.renderRectangleEx(x, y, w, h, 0);
}

pub const OutputSize = struct { width: c_int, height: c_int };
pub fn getOutputSize(self: Self) OutputSize {
    var wsize = self.window.getSize();
    return OutputSize{
        .width = wsize.width,
        .height = wsize.height,
    };
}

const image = [_]u8{
    // 1
    0,0,0,
    255,255,255,
    0,0,0,
    255,255,255,
    // 2
    255,255,255,
    0,0,0,
    255,255,255,
    0,0,0,
    // 1
    0,0,0,
    255,255,255,
    0,0,0,
    255,255,255,
    // 2
    255,255,255,
    0,0,0,
    255,255,255,
    0,0,0,
};

const Text = struct {
    texture: gl.Texture,

    pub fn init(text: [:0]const u8) Text {
        var tex = gl.genTexture();
        //defer gl.Texture.delete(tex);
        gl.bindTexture(tex, .@"2d");

        var font = sdl.ttf.openFont("res/fonts/MarginaliaRegular-8XlZ.ttf", 200) catch unreachable;
        //defer font.close();
        var surface = font.renderTextBlended(text, sdl.Color.white) catch unreachable;
        const width = @intCast(usize, surface.ptr.w);
        const height = @intCast(usize, surface.ptr.h);
        var newsurf = sdl.c.SDL_ConvertSurfaceFormat(surface.ptr, @enumToInt(sdl.PixelFormatEnum.argb8888), 0) orelse unreachable;
        const data = @ptrCast([*]const u8, newsurf.pixels);
        //defer surface.destroy();

        gl.textureImage2D(.@"2d", 0, .rgba, width, height, .rgba, .unsigned_byte, data);

        gl.texParameter(.@"2d", .wrap_s, .clamp_to_edge);
        gl.texParameter(.@"2d", .wrap_r, .clamp_to_edge);
        gl.texParameter(.@"2d", .min_filter, .linear);
        gl.texParameter(.@"2d", .mag_filter, .linear);

        return Text {
            .texture = tex,
        };
    }

    pub fn deinit(self: *Text) void {
        _ = self;
    }
};
